class VariableNotFoundError(Exception):
    """Raised when a variable could not be found in the context of a
    template.
    """


class Context(dict):
    def __init__(self, filename, **context):
        super(Context, self).__init__(**context)

        self.filename = filename

    def update(self, **context):
        super(Context, self).update(**context)
        return self

    def lookup(self, key, default=None):
        """Look up a variable in the current context.

        :raises VariableNotFoundError:
          When a variable could not be found.
        :param str key:
          The variable to look up.
        :param default:
          The default value to return if the variable doesn't exist.
        :returns:
          The value of the variable.
        """
        if key not in self:
            if default is not None:
                return default

            raise VariableNotFoundError("variable '{}' not found in '{}'".format(
                key, self.filename
            ))

        return self[key]



from collections import namedtuple

from .tokenizer import TokenKind

Node = namedtuple("Node", ("kind", "token", "children"))
"""An abstract data type representing a template node.
"""


class TemplateSyntaxError(SyntaxError):
    """Raised when a syntax error was encountered during parsing.
    """


class PopBlock(Exception):
    """Raised when the parse stack for the current block needs to be
    popped (i.e. when an end tag has been reached).
    """


def syntax_error(token, message):
    """Raise a :class:`.TemplateSyntaxError`.

    :param Token token:
      The :class:`.Token` to reference in the error message. This is
      used to determine which file, line and column the error occurred
      at.
    :param str message:
      The error message.
    """
    return TemplateSyntaxError("{filename}:{line}:{column}: {message}".format(
        filename=token.filename,
        line=token.line, column=token.column,
        message=message
    ))


def parse_block(token, tokens):
    """Parse a nested block.

    :raises TemplateSyntaxError:
      When the token iterator has been fully consumed but a
      :class:`.PopBlock` hasn't been raised.
    :param Token token:
      The base :class:`.Token` that represents this block.
    :param Iterator tokens:
      A :class:`.Token` iterator. This is consumed until a
      :class:`.PopBlock` is raised.
    :returns:
      A block :class:`.Node`.
    """
    children = []
    try:
        for child in parse(tokens, pop=True):
            children.append(child)
    except PopBlock:
        return Node(token.kind, token, children)
    else:
        raise syntax_error(token, "missing {end}")


def parse(tokens, pop=False):
    """Parse a stream of :class:`Tokens<.Token>` into a stream of
    :class:`Nodes<.Node>`.

    :raises TemplateSyntaxError:
      When an unexpected :class:`.Token` was encountered.
    :param Iterator tokens:
      The stream of :class:`Tokens<.Token>` to consume.
    :param bool pop:
      Whether or not to pop on {end}.
    :returns:
      A :class:`.Node` generator.
    """
    for token in tokens:
        if token.kind in (
                TokenKind.when,
                TokenKind.unless,
                TokenKind.foreach,
                TokenKind.block,
                TokenKind.extend_block
        ):
            yield parse_block(token, tokens)

        elif token.kind in (
                TokenKind.chunk,
                TokenKind.raw,
                TokenKind.var,
                TokenKind.extend,
                TokenKind.include
        ):
            yield Node(token.kind, token, None)

        elif token.kind == TokenKind.end and pop:
            raise PopBlock

        else:
            raise syntax_error(token, "unexpected '{}'".format(token.kind))



import re

from collections import namedtuple
from functools import partial

BASE_VARIABLE_RE = r"([a-zA-Z_][a-zA-Z0-9]*)"
VARIABLE_RE = r"{}([^}}]+)?".format(BASE_VARIABLE_RE)
FILENAME_RE = r"([^}]+)"
TOKENS = (
    ("newline", r"\n"),
    ("chunk", r"[^{]+"),
    ("end", r"{end}"),
    ("escaped_bracket", r"{{"),
    ("raw", r"{{\@{var}}}".format(var=VARIABLE_RE)),
    ("var", r"{{\+{var}}}".format(var=VARIABLE_RE)),
    ("when", r"{{\?{var}}}".format(var=VARIABLE_RE)),
    ("unless", r"{{\!{var}}}".format(var=VARIABLE_RE)),
    ("foreach", r"{{\*{var}}}".format(var=VARIABLE_RE)),
    ("block", r"{{#{var}}}".format(var=BASE_VARIABLE_RE)),
    ("extend_block", r"{{\^#{var}}}".format(var=BASE_VARIABLE_RE)),
    ("extend", r"{{\^{filename}}}".format(filename=FILENAME_RE)),
    ("include", r"{{>{filename}}}".format(filename=FILENAME_RE)),
)
TOKENIZER = "|".join("(?P<{}>{})".format(k, e) for k, e in TOKENS)
TOKENIZER = re.compile(TOKENIZER)

Token = namedtuple("Token", ("kind", "value", "filename", "line", "column"))
TokenKind = type("TokenKind", (object,), {})
for kind, _ in TOKENS:
    setattr(TokenKind, kind, kind)


def tokenize(filename, template):
    """Tokenize a template.

    :raises IOError:
      When the file couldn't be read.
    :param str template:
      The contents of a template.
    :returns:
      A :class:`.Token` generator.
    """
    chunks = []  # Keep track of and merge consecutive chunks.
    chunk_like = TokenKind.newline, TokenKind.chunk, TokenKind.escaped_bracket
    line, column = 1, 0
    for match in TOKENIZER.finditer(template):
        kind = match.lastgroup
        value = match.group(kind)
        token = partial(Token, filename=filename, line=line, column=column)

        if kind in chunk_like:
            if not chunks:
                cline, ccolumn = line, column

            if kind == TokenKind.escaped_bracket:
                chunks.append(value[0])
            else:
                chunks.append(value)

            newlines = value.count("\n")
            if newlines:
                column = 0

            line += newlines
            column += len(value) - value.rfind("\n") - 1
        else:
            if chunks:
                yield Token(TokenKind.chunk, "".join(chunks), filename, cline, ccolumn)
                chunks = []

            if kind == TokenKind.end:
                yield token(kind=kind, value=value)

            elif kind in (TokenKind.include, TokenKind.extend):
                name = match.group(match.lastindex + 1)
                yield token(kind=kind, value=name)

            elif kind in (TokenKind.block, TokenKind.extend_block):
                name = match.group(match.lastindex + 1)
                yield token(kind=kind, value=name)

            else:
                variable = match.group(match.lastindex + 1)
                suffix = match.group(match.lastindex + 2)
                yield token(kind=kind, value=(variable, suffix))

            column += len(value)

    if chunks:
        yield Token(TokenKind.chunk, "".join(chunks), filename, cline, ccolumn)



from .environment import Environment  # noqa

# Re-export exceptions:
from .context import VariableNotFoundError  # noqa
from .environment import TemplateNotFoundError  # noqa
from .parser import TemplateSyntaxError  # noqa



from collections import namedtuple

from .context import Context


class Template(namedtuple("Template", ("filename", "name", "function", "blocks"))):
    """Abstract data type representing a compiled template.

    :param str filename:
      The template's filename.
    :param str name:
      The template function's name.
    :param callable function:
      The compiled template.
    :param Context blocks:
      The template's block context.
    """
    def render(self, **context):
        context = Context(self.filename, **context)
        return self.function(self.blocks, context)



"Here be dragons."

from .context import Context
from .template import Template
from .tokenizer import TokenKind

LOOKUP = "context.lookup({!r}){}"
RAW = "str({})".format(LOOKUP)
VAR = "escape({})".format(RAW)
WHEN = "{func}(blocks, context) if {var} else ''"
UNLESS = "{func}(blocks, context) if not {var} else ''"
FOREACH = "''.join({func}(blocks, context.update(even=(i % 2 != 0), item=item)) for i, item in enumerate({var}))"
BLOCK = "blocks.lookup({var!r}, {func})(blocks, context) if blocks else {func}(blocks, context)"
EXTEND = INCLUDE = "{func}(blocks, context)"
FUNCTION = """\
def {name}(blocks, context):
  return "".join([{nodes}])\
"""


def gennested(env, var, node):
    token = node.token
    name = ":".join((token.filename, token.kind, var, str(token.line), str(token.column)))
    return env._compile(name, node.children)


def gennode(env, node, blocks):
    if node.kind not in (
            TokenKind.chunk,
            TokenKind.end,
            TokenKind.include,
            TokenKind.block,
            TokenKind.extend_block,
            TokenKind.extend,
    ):
        var, suffix = node.token.value
        var = var, suffix or ""

    if node.kind == TokenKind.chunk:
        return repr(node.token.value)

    elif node.kind == TokenKind.raw:
        return RAW.format(*var)

    elif node.kind == TokenKind.var:
        return VAR.format(*var)

    elif node.kind == TokenKind.include:
        template = env._load(node.token.value)
        return INCLUDE.format(func=template.name)

    elif node.kind == TokenKind.block:
        var = node.token.value
        template = gennested(env, var, node)
        return BLOCK.format(var=var, func=template.name)

    elif node.kind == TokenKind.extend_block:
        var = node.token.value
        template = gennested(env, var, node)
        blocks[var] = template.function
        return None

    elif node.kind == TokenKind.extend:
        template = env._load(node.token.value)
        return EXTEND.format(func=template.name)

    elif node.kind in (TokenKind.when, TokenKind.unless, TokenKind.foreach):
        template = gennested(env, "".join(var), node)
        itername = template.name
        return {
            TokenKind.when: WHEN,
            TokenKind.unless: UNLESS,
            TokenKind.foreach: FOREACH
        }[node.kind].format(func=itername, var=LOOKUP.format(*var))


def genfunc(env, filename, name, nodes):
    blocks = Context(filename)
    nodes = ", ".join(filter(lambda x: x is not None, (gennode(env, x, blocks) for x in nodes)))
    source = FUNCTION.format(name=name, nodes=nodes)
    exec(source, env.context, env.context)

    function = env.context[name]
    function.__source__ = source
    env.template_cache[name] = template = Template(filename, name, function, blocks)
    return template



from setuptools import setup

setup(
    name="pymold",
    version="0.1.0.2",
    description="A simple, fast template engine for Python.",
    long_description="https://github.com/Bogdanp/mold",
    packages=[
        "mold",
    ],

    install_requires=[
    ],

    author="Bogdan Popa",
    author_email="popa.bogdanp@gmail.com",
    url="https://github.com/Bogdanp/mold",
    keywords=["web"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
    ]
)



import hashlib
import os

from xml.sax.saxutils import escape

from .codegen import genfunc
from .context import Context
from .parser import parse
from .tokenizer import tokenize


class TemplateNotFoundError(Exception):
    """Raised when a template could not be found under any of the
    provided search paths.
    """


def normalize_name(filename):
    """Map a template file name to a template function name.

    :param str filename:
      The file name to map.
    :returns:
      The mapped function name.
    """
    return "_template_" + hashlib.md5(filename.encode("utf-8")).hexdigest()


class Environment(object):
    r"""A template environment.

    This is the primary mechanism through which templates are loaded
    from disk.

    :param list paths:
      A list of search paths. These are searched in order and the
      first match is always compiled and returned.
    :param \**context:
      Any names you would like to make avaialble in the global context
      of every compiled template.
    """

    def __init__(self, paths, **context):
        if isinstance(paths, str):
            paths = (paths,)

        self.paths_ = tuple(paths)
        self.context = dict(Context=Context, escape=escape, **context)
        self.template_cache = {}
        self.filename_cache = {}

    @property
    def paths(self):
        """Accessor for the search paths.

        :returns:
          A tuple of search paths.
        """
        return self.paths_

    @paths.setter
    def paths(self, value):
        """Setter for the search paths.

        Ensures that the underlying value is a tuple and busts the
        filename cache on every set action.

        :param tuple value:
          A tuple of search paths.
        """
        self.filename_cache = {}
        self.paths_ = tuple(value)

    def _load(self, filename):
        if filename not in self.filename_cache:
            for path in self.paths:
                filepath = os.path.join(path, filename)
                if os.path.isfile(filepath):
                    self.filename_cache[filename] = (filepath, normalize_name(filepath))
                    break
            else:
                raise TemplateNotFoundError("template '{}' not found".format(filename))

        filename, name = self.filename_cache[filename]
        if name in self.template_cache:
            return self.template_cache[name]

        with open(filename) as f:
            contents = f.read()

        return self._compile(filename, parse(tokenize(filename, contents)))

    def _compile(self, filename, nodes):
        name = normalize_name(filename)
        return genfunc(self, filename, name, nodes)

    def get_template(self, filename):
        """Load a template from disk.

        :raises TemplateNotFoundError:
          When the template could not be found.
        :param str filename:
          The name of the template to load.
        :return:
          A template.
        """
        return self._load(filename)

    def render(self, filename, **context):
        """Render a template.

        :raises VariableNotFoundError:
          If the template references a variable that wasn't provided
          in the context.
        :raises TemplateNotFoundError:
          If the template could not be found in the search path.
        :param str filename:
          The name of the template to render.
        :param \**context:
          The context with which to render the template.
        :return:
          A string representing the rendered template.
        """
        return self._load(filename).render(**context)
